use std::{
    cell::RefCell,
    rc::{Rc, Weak},
};

use im::hashmap::*;
use wasm_bindgen::prelude::*;
use web_sys::{Element, Node};
use yew::prelude::*;

use crate::lineengine::line::{Readline, ReadlineEvent};

#[wasm_bindgen]
extern "C" {
    #[derive(PartialEq)]
    pub type Terminal;
    // Sadly, can't do type parameters with wasm_bindgen.
    pub type IEventString;
    pub type IEventDims;
    pub type IDisposable;

    #[wasm_bindgen(constructor)]
    fn new() -> Terminal;
    #[wasm_bindgen(method)]
    fn open(this: &Terminal, element: &Element);
    #[wasm_bindgen(method)]
    pub fn write(this: &Terminal, data: &str);
    // Todo: Can we do this with interfaces somehow?
    #[wasm_bindgen(method)]
    fn loadAddon(this: &Terminal, addon: &FitAddon);

    #[wasm_bindgen(method)]
    fn onData(this: &Terminal, listener: &Closure<dyn FnMut(String)>) -> IDisposable;
    #[wasm_bindgen(method)]
    fn onResize(this: &Terminal, listener: &Closure<dyn FnMut(Dims)>) -> IDisposable;

    #[wasm_bindgen(method, getter)]
    fn rows(this: &Terminal) -> u16;
    #[wasm_bindgen(method, getter)]
    fn cols(this: &Terminal) -> u16;

    pub type Dims;
    #[wasm_bindgen(method, getter)]
    fn rows(this: &Dims) -> u16;
    #[wasm_bindgen(method, getter)]
    fn cols(this: &Dims) -> u16;

    #[wasm_bindgen(method, setter)]
    fn set_listener(this: &IEventString, handler: &Closure<dyn FnMut(String)>) -> IDisposable;
    #[wasm_bindgen(method, setter)]
    fn set_listener(this: &IEventDims, handler: &Closure<dyn FnMut(Dims)>) -> IDisposable;
    #[wasm_bindgen(method)]
    fn dispose(this: &IDisposable);

    #[derive(PartialEq)]
    pub type FitAddon;
}

#[wasm_bindgen(js_namespace=FitAddon)]
extern "C" {
    #[wasm_bindgen(constructor)]
    fn new() -> FitAddon;
    #[wasm_bindgen(method)]
    fn fit(this: &FitAddon);
}

#[derive(Eq, Ord, Hash, PartialEq, PartialOrd, Clone)]
pub struct TermFrame(pub u64);

#[derive(Properties)]
pub struct TermFrameData {
    pub id: TermFrame,
    pub term: Terminal,
    pub fit: FitAddon,
    pub node: Node,
    pub readline: RefCell<Readline>,
    pub retained_closures: RefCell<Option<(Closure<dyn FnMut(String)>, Closure<dyn FnMut(Dims)>)>>,
}

impl PartialEq for TermFrameData {
    fn eq(&self, other: &Self) -> bool {
        // Only the ID matters, the rest are just reference data.
        self.id == other.id
    }
}

pub type RegisteredTermFrames = HashMap<TermFrame, Rc<TermFrameData>>;
#[derive(Properties, PartialEq, Clone)]
pub struct RegisteredTermFrameLens {
    pub get: Callback<(), Rc<RegisteredTermFrames>>,
    pub set: Callback<Rc<RegisteredTermFrames>, ()>,
}

fn get_or_make_term_frame(
    frame: &TermFrame,
    frames: &RegisteredTermFrameLens,
    handler: &Callback<(TermFrame, String), ()>,
) -> Rc<TermFrameData> {
    if let Some(tfd) = frames.get.emit(()).get(frame) {
        return tfd.clone();
    }

    let mut new_frames: RegisteredTermFrames = (*frames.get.emit(())).clone();
    let term = Terminal::new();
    let fit = FitAddon::new();
    let element = web_sys::window()
        .and_then(|w| w.document())
        .and_then(|d| d.create_element("div").ok())
        .expect("Can create element for term");
    element.set_class_name("hterminal");
    term.open(&element);
    term.loadAddon(&fit);
    fit.fit();
    for i in 0..100 {
        term.write(&format!("{} Hello world\r\n", i));
    }
    let term_for_readline: Terminal = Terminal { obj: term.clone() };
    let initial_size = (term.cols(), term.rows());
    let new_data: Rc<TermFrameData> = TermFrameData {
        id: frame.clone(),
        term: Terminal { obj: term.clone() },
        fit,
        node: element.into(),
        readline: Readline::new(
            "".to_owned(),
            Box::new(move |dat| {
                term_for_readline
                    .write(std::str::from_utf8(dat).expect("readline tried to emit invalid UTF-8"))
            }),
            initial_size,
        )
        .into(),
        retained_closures: RefCell::new(None),
    }
    .into();

    let data_for_resize: Weak<TermFrameData> = Rc::downgrade(&new_data);
    let resize_closure = Closure::new(move |dims: Dims| match Weak::upgrade(&data_for_resize) {
        None => {}
        Some(r) => match r.readline.try_borrow_mut() {
            Err(_) => {}
            Ok(mut v) => v.handle_resize((dims.cols(), dims.rows())).unwrap_or(()),
        },
    });
    term.onResize(&resize_closure);

    let cloned_handler = (*handler).clone();
    let emit_frame = frame.clone();
    let data_for_on_data: Weak<TermFrameData> = Rc::downgrade(&new_data);

    let data_closure = Closure::new(move |d: String| match Weak::upgrade(&data_for_on_data) {
        None => {}
        Some(r) => match r.readline.try_borrow_mut() {
            Err(_) => {}
            Ok(mut v) => {
                for ev in v.readline(d.as_bytes()).expect("Readline failed") {
                    match ev {
                        ReadlineEvent::Line(l) => cloned_handler.emit((emit_frame.clone(), l)),
                        _ => {}
                    }
                }
            }
        },
    });

    term.onData(&data_closure);
    new_data
        .retained_closures
        .replace(Some((data_closure, resize_closure)));

    new_frames.insert(frame.clone(), new_data.clone());
    web_sys::console::log_2(
        &"Setting frames to have length: ".into(),
        &new_frames.iter().count().into(),
    );
    frames.set.emit(new_frames.into());
    new_data
}

#[derive(Properties, PartialEq)]
pub struct TermViewProps {
    pub terminal: TermFrame,
    pub frames: RegisteredTermFrameLens,
    pub handler: Callback<(TermFrame, String), ()>,
}

#[function_component(TermView)]
pub fn term_view(props: &TermViewProps) -> Html {
    let term = get_or_make_term_frame(&props.terminal, &props.frames, &props.handler);
    term.fit.fit();
    html! {
        {Html::VRef(term.node.clone())}
    }
}

pub fn echo_to_term_frame(
    frames: &RegisteredTermFrameLens,
    frame_id: &TermFrame,
    message: &str,
) -> Result<(), &'static str> {
    let frame_val = frames.get.emit(());
    let frame: &TermFrameData = frame_val.get(frame_id).ok_or_else(|| {
        web_sys::console::log_3(
            &"Attempt to echo to frame that doesn't exist.".into(),
            &frame_id.0.into(),
            &frame_val.iter().count().into(),
        );
        "Attempt to echo to frame that doesn't exist."
    })?;
    frame.term.write(message);
    Ok(())
}
